@@ -1,5 +1,5 @@
 /* Command processing for GNU Make.
-Copyright (C) 1988-2022 Free Software Foundation, Inc.
+Copyright (C) 1988-2023 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify it under the
@@ -12,10 +12,11 @@ WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along with
-this program.  If not, see <http://www.gnu.org/licenses/>.  */
+this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 #include "makeint.h"
 #include "filedef.h"
+#include "os.h"
 #include "dep.h"
 #include "variable.h"
 #include "job.h"
@@ -321,14 +322,12 @@ set_file_variables (struct file *file, const char *stem)
 void
 chop_commands (struct commands *cmds)
 {
-  unsigned int nlines;
-  unsigned short idx;
+  unsigned short nlines;
+  unsigned short i;
   char **lines;
 
-  /* If we don't have any commands,
-     or we already parsed them, never mind.  */
-
-  if (!cmds || cmds->command_lines != 0)
+  /* If we don't have any commands, or we already parsed them, never mind.  */
+  if (!cmds || cmds->command_lines != NULL)
     return;
 
   /* Chop CMDS->commands up into lines in CMDS->command_lines.  */
@@ -347,25 +346,27 @@ chop_commands (struct commands *cmds)
     }
   else
     {
-      const char *p;
+      const char *p = cmds->commands;
+      size_t max = 5;
 
-      nlines = 5;
-      lines = xmalloc (nlines * sizeof (char *));
-      idx = 0;
-      p = cmds->commands;
+      nlines = 0;
+      lines = xmalloc (max * sizeof (char *));
       while (*p != '\0')
         {
           const char *end = p;
         find_end:;
           end = strchr (end, '\n');
-          if (end == 0)
+          if (end == NULL)
             end = p + strlen (p);
           else if (end > p && end[-1] == '\\')
             {
               int backslash = 1;
-              const char *b;
-              for (b = end - 2; b >= p && *b == '\\'; --b)
-                backslash = !backslash;
+              if (end > p + 1)
+                {
+                  const char *b;
+                  for (b = end - 2; b >= p && *b == '\\'; --b)
+                    backslash = !backslash;
+                }
               if (backslash)
                 {
                   ++end;
@@ -373,40 +374,36 @@ chop_commands (struct commands *cmds)
                 }
             }
 
-          if (idx == nlines)
+          if (nlines == USHRT_MAX)
+            ON (fatal, &cmds->fileinfo,
+                _("Recipe has too many lines (limit %hu)"), nlines);
+
+          if (nlines == max)
             {
-              nlines += 2;
-              lines = xrealloc (lines, nlines * sizeof (char *));
+              max += 2;
+              lines = xrealloc (lines, max * sizeof (char *));
             }
-          lines[idx++] = xstrndup (p, (size_t) (end - p));
+
+          lines[nlines++] = xstrndup (p, (size_t) (end - p));
           p = end;
           if (*p != '\0')
             ++p;
         }
-
-      if (idx != nlines)
-        {
-          nlines = idx;
-          lines = xrealloc (lines, nlines * sizeof (char *));
-        }
     }
 
   /* Finally, set the corresponding CMDS->lines_flags elements and the
      CMDS->any_recurse flag.  */
 
-  if (nlines > USHRT_MAX)
-    ON (fatal, &cmds->fileinfo, _("Recipe has too many lines (%ud)"), nlines);
-
-  cmds->ncommand_lines = (unsigned short)nlines;
+  cmds->ncommand_lines = nlines;
   cmds->command_lines = lines;
 
   cmds->any_recurse = 0;
   cmds->lines_flags = xmalloc (nlines);
 
-  for (idx = 0; idx < nlines; ++idx)
+  for (i = 0; i < nlines; ++i)
     {
       unsigned char flags = 0;
-      const char *p = lines[idx];
+      const char *p = lines[i];
 
       while (ISBLANK (*p) || *p == '-' || *p == '@' || *p == '+')
         switch (*(p++))
@@ -423,12 +420,12 @@ chop_commands (struct commands *cmds)
           }
 
       /* If no explicit '+' was given, look for MAKE variable references.  */
-      if (!(flags & COMMANDS_RECURSE)
+      if (! ANY_SET (flags, COMMANDS_RECURSE)
           && (strstr (p, "$(MAKE)") != 0 || strstr (p, "${MAKE}") != 0))
         flags |= COMMANDS_RECURSE;
 
-      cmds->lines_flags[idx] = flags;
-      cmds->any_recurse |= flags & COMMANDS_RECURSE ? 1 : 0;
+      cmds->lines_flags[i] = flags;
+      cmds->any_recurse |= ANY_SET (flags, COMMANDS_RECURSE) ? 1 : 0;
     }
 }
 
@@ -479,7 +476,7 @@ execute_file_commands (struct file *file)
 /* This is set while we are inside fatal_error_signal,
    so things can avoid nonreentrant operations.  */
 
-int handling_fatal_signal = 0;
+volatile sig_atomic_t handling_fatal_signal = 0;
 
 /* Handle fatal signals.  */
 
@@ -532,6 +529,10 @@ fatal_error_signal (int sig)
      It is blocked now while we run this handler.  */
   signal (sig, SIG_DFL);
 
+  temp_stdin_unlink ();
+  osync_clear ();
+  jobserver_clear ();
+
   /* A termination signal won't be sent to the entire
      process group, but it means we want to kill the children.  */
 
