@@ -1,5 +1,5 @@
 /* Target file management for GNU Make.
-Copyright (C) 1988-2022 Free Software Foundation, Inc.
+Copyright (C) 1988-2023 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify it under the
@@ -12,7 +12,7 @@ WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along with
-this program.  If not, see <http://www.gnu.org/licenses/>.  */
+this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 #include "makeint.h"
 
@@ -63,9 +63,6 @@ static struct hash_table files;
 /* Whether or not .SECONDARY with no prerequisites was given.  */
 static int all_secondary = 0;
 
-/* Whether or not .NOTINTERMEDIATE with no prerequisites was given.  */
-static int no_intermediates = 0;
-
 /* Access the hash table of all file records.
    lookup_file  given a name, return the struct file * for that name,
                 or nil if there is none.
@@ -108,20 +105,10 @@ lookup_file (const char *name)
   while (name[0] == '<' && name[1] == '>' && name[2] != '\0')
       name += 2;
 #endif
-  while (name[0] == '.'
-#ifdef HAVE_DOS_PATHS
-         && (name[1] == '/' || name[1] == '\\')
-#else
-         && name[1] == '/'
-#endif
-         && name[2] != '\0')
+  while (name[0] == '.' && ISDIRSEP (name[1]) && name[2] != '\0')
     {
       name += 2;
-      while (*name == '/'
-#ifdef HAVE_DOS_PATHS
-             || *name == '\\'
-#endif
-             )
+      while (ISDIRSEP (*name))
         /* Skip following slashes: ".//foo" is "foo", not "/foo".  */
         ++name;
     }
@@ -378,7 +365,7 @@ remove_intermediates (int sig)
   int doneany = 0;
 
   /* If there's no way we will ever remove anything anyway, punt early.  */
-  if (question_flag || touch_flag || all_secondary)
+  if (question_flag || touch_flag || all_secondary || no_intermediates)
     return;
 
   if (sig && just_print_flag)
@@ -741,7 +728,7 @@ snap_file (const void *item, void *arg)
   /* If .NOTINTERMEDIATE is set with no deps, mark all targets as
      notintermediate, unless the target is a prereq of .INTERMEDIATE.  */
   if (no_intermediates && !f->intermediate && !f->secondary)
-      f->notintermediate = 1;
+    f->notintermediate = 1;
 
   /* If .EXTRA_PREREQS is set, add them as ignored by automatic variables.  */
   if (f->variables)
@@ -820,7 +807,7 @@ snap_deps (void)
     else
       no_intermediates = 1;
 
-  /* The same file connot be both .INTERMEDIATE and .NOTINTERMEDIATE.
+  /* The same file cannot be both .INTERMEDIATE and .NOTINTERMEDIATE.
      However, it is possible for a file to be .INTERMEDIATE and also match a
      .NOTINTERMEDIATE pattern.  In that case, the intermediate file has
      priority over the notintermediate pattern.  This priority is enforced by
@@ -1018,13 +1005,16 @@ file_timestamp_sprintf (char *p, FILE_TIMESTAMP ts)
   struct tm *tm = localtime (&t);
 
   if (tm)
-    sprintf (p, "%04d-%02d-%02d %02d:%02d:%02d",
-             tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
-             tm->tm_hour, tm->tm_min, tm->tm_sec);
+    {
+      intmax_t year = tm->tm_year;
+      sprintf (p, "%04" PRIdMAX "-%02d-%02d %02d:%02d:%02d",
+               year + 1900, tm->tm_mon + 1, tm->tm_mday,
+               tm->tm_hour, tm->tm_min, tm->tm_sec);
+    }
   else if (t < 0)
-    sprintf (p, "%ld", (long) t);
+    sprintf (p, "%" PRIdMAX, (intmax_t) t);
   else
-    sprintf (p, "%lu", (unsigned long) t);
+    sprintf (p, "%" PRIuMAX, (uintmax_t) t);
   p += strlen (p);
 
   /* Append nanoseconds as a fraction, but remove trailing zeros.  We don't
