@@ -83,6 +83,12 @@ extern int errno;
 #endif
 
 /* Define macros specifying which OS we are building for.  */
+#if __gnu_hurd__
+# define MK_OS_HURD 1
+#endif
+#if __CYGWIN__
+# define MK_OS_CYGWIN 1
+#endif
 #if defined(__MVS__)
 # define MK_OS_ZOS 1
 #endif
@@ -390,7 +396,9 @@ extern int unixy_shell;
 # endif
 
 /* Include only the minimal stuff from windows.h.   */
-# define WIN32_LEAN_AND_MEAN
+# ifndef WIN32_LEAN_AND_MEAN
+#  define WIN32_LEAN_AND_MEAN
+# endif
 #endif  /* WINDOWS32 */
 
 /* ALL_SET() evaluates the second argument twice.  */
@@ -742,6 +750,14 @@ extern char cmd_prefix;
 
 extern unsigned int no_intermediates;
 
+#if HAVE_MKFIFO
+/* It seems that mkfifo() is not working correctly, or at least not the way
+   GNU make wants it to work, on GNU/Hurd and Cygwin so don't use it there.  */
+# if !defined(JOBSERVER_USE_FIFO) && !MK_OS_HURD && !MK_OS_CYGWIN
+#  define JOBSERVER_USE_FIFO 1
+# endif
+#endif
+
 #define JOBSERVER_AUTH_OPT      "jobserver-auth"
 
 extern char *jobserver_auth;
